# DO NOT EDIT! This file is automatically generated
import marshmallow
import marshmallow_enum

from commercetools import helpers, types
from commercetools._schemas._common import LocalizedStringField

__all__ = [
    "AccessDeniedErrorSchema",
    "ConcurrentModificationErrorSchema",
    "DiscountCodeNonApplicableErrorSchema",
    "DuplicateAttributeValueErrorSchema",
    "DuplicateAttributeValuesErrorSchema",
    "DuplicateFieldErrorSchema",
    "DuplicateFieldWithConflictingResourceErrorSchema",
    "DuplicatePriceScopeErrorSchema",
    "DuplicateVariantValuesErrorSchema",
    "EnumValueIsUsedErrorSchema",
    "ErrorByExtensionSchema",
    "ErrorObjectSchema",
    "ErrorResponseSchema",
    "ExtensionBadResponseErrorSchema",
    "ExtensionNoResponseErrorSchema",
    "ExtensionUpdateActionsFailedErrorSchema",
    "InsufficientScopeErrorSchema",
    "InvalidCredentialsErrorSchema",
    "InvalidCurrentPasswordErrorSchema",
    "InvalidFieldErrorSchema",
    "InvalidInputErrorSchema",
    "InvalidItemShippingDetailsErrorSchema",
    "InvalidJsonInputErrorSchema",
    "InvalidOperationErrorSchema",
    "InvalidSubjectErrorSchema",
    "InvalidTokenErrorSchema",
    "LanguageUsedInStoresErrorSchema",
    "MatchingPriceNotFoundErrorSchema",
    "MissingTaxRateForCountryErrorSchema",
    "NoMatchingProductDiscountFoundErrorSchema",
    "OutOfStockErrorSchema",
    "PriceChangedErrorSchema",
    "ReferenceExistsErrorSchema",
    "ReferencedResourceNotFoundErrorSchema",
    "RequiredFieldErrorSchema",
    "ResourceNotFoundErrorSchema",
    "ShippingMethodDoesNotMatchCartErrorSchema",
    "VariantValuesSchema",
]


class ErrorByExtensionSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ErrorByExtension`."""

    id = marshmallow.fields.String(allow_none=True)
    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ErrorByExtension(**data)


class ErrorObjectSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ErrorObject`."""

    code = marshmallow.fields.String(allow_none=True)
    message = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ErrorObject(**data)


class ErrorResponseSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.ErrorResponse`."""

    status_code = marshmallow.fields.Integer(allow_none=True, data_key="statusCode")
    message = marshmallow.fields.String(allow_none=True)
    error = marshmallow.fields.String(allow_none=True, missing=None)
    error_description = marshmallow.fields.String(allow_none=True, missing=None)
    errors = marshmallow.fields.List(
        helpers.Discriminator(
            discriminator_field=("code", "code"),
            discriminator_schemas={
                "access_denied": "commercetools._schemas._error.AccessDeniedErrorSchema",
                "ConcurrentModification": "commercetools._schemas._error.ConcurrentModificationErrorSchema",
                "DiscountCodeNonApplicable": "commercetools._schemas._error.DiscountCodeNonApplicableErrorSchema",
                "DuplicateAttributeValue": "commercetools._schemas._error.DuplicateAttributeValueErrorSchema",
                "DuplicateAttributeValues": "commercetools._schemas._error.DuplicateAttributeValuesErrorSchema",
                "DuplicateField": "commercetools._schemas._error.DuplicateFieldErrorSchema",
                "DuplicateFieldWithConflictingResource": "commercetools._schemas._error.DuplicateFieldWithConflictingResourceErrorSchema",
                "DuplicatePriceScope": "commercetools._schemas._error.DuplicatePriceScopeErrorSchema",
                "DuplicateVariantValues": "commercetools._schemas._error.DuplicateVariantValuesErrorSchema",
                "EnumValueIsUsed": "commercetools._schemas._error.EnumValueIsUsedErrorSchema",
                "ExtensionBadResponse": "commercetools._schemas._error.ExtensionBadResponseErrorSchema",
                "ExtensionNoResponse": "commercetools._schemas._error.ExtensionNoResponseErrorSchema",
                "ExtensionUpdateActionsFailed": "commercetools._schemas._error.ExtensionUpdateActionsFailedErrorSchema",
                "insufficient_scope": "commercetools._schemas._error.InsufficientScopeErrorSchema",
                "InvalidCredentials": "commercetools._schemas._error.InvalidCredentialsErrorSchema",
                "InvalidCurrentPassword": "commercetools._schemas._error.InvalidCurrentPasswordErrorSchema",
                "InvalidField": "commercetools._schemas._error.InvalidFieldErrorSchema",
                "InvalidInput": "commercetools._schemas._error.InvalidInputErrorSchema",
                "InvalidItemShippingDetails": "commercetools._schemas._error.InvalidItemShippingDetailsErrorSchema",
                "InvalidJsonInput": "commercetools._schemas._error.InvalidJsonInputErrorSchema",
                "InvalidOperation": "commercetools._schemas._error.InvalidOperationErrorSchema",
                "InvalidSubject": "commercetools._schemas._error.InvalidSubjectErrorSchema",
                "invalid_token": "commercetools._schemas._error.InvalidTokenErrorSchema",
                "LanguageUsedInStores": "commercetools._schemas._error.LanguageUsedInStoresErrorSchema",
                "MatchingPriceNotFound": "commercetools._schemas._error.MatchingPriceNotFoundErrorSchema",
                "MissingTaxRateForCountry": "commercetools._schemas._error.MissingTaxRateForCountryErrorSchema",
                "NoMatchingProductDiscountFound": "commercetools._schemas._error.NoMatchingProductDiscountFoundErrorSchema",
                "OutOfStock": "commercetools._schemas._error.OutOfStockErrorSchema",
                "PriceChanged": "commercetools._schemas._error.PriceChangedErrorSchema",
                "ReferenceExists": "commercetools._schemas._error.ReferenceExistsErrorSchema",
                "ReferencedResourceNotFound": "commercetools._schemas._error.ReferencedResourceNotFoundErrorSchema",
                "RequiredField": "commercetools._schemas._error.RequiredFieldErrorSchema",
                "ResourceNotFound": "commercetools._schemas._error.ResourceNotFoundErrorSchema",
                "ShippingMethodDoesNotMatchCart": "commercetools._schemas._error.ShippingMethodDoesNotMatchCartErrorSchema",
            },
            unknown=marshmallow.EXCLUDE,
            allow_none=True,
        ),
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.ErrorResponse(**data)


class VariantValuesSchema(marshmallow.Schema):
    """Marshmallow schema for :class:`commercetools.types.VariantValues`."""

    sku = marshmallow.fields.String(allow_none=True, missing=None)
    prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceDraftSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )
    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        return types.VariantValues(**data)


class AccessDeniedErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.AccessDeniedError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.AccessDeniedError(**data)


class ConcurrentModificationErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ConcurrentModificationError`."""

    current_version = marshmallow.fields.Integer(
        allow_none=True, missing=None, data_key="currentVersion"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ConcurrentModificationError(**data)


class DiscountCodeNonApplicableErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DiscountCodeNonApplicableError`."""

    discount_code = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="discountCode"
    )
    reason = marshmallow.fields.String(allow_none=True, missing=None)
    dicount_code_id = marshmallow.fields.String(
        allow_none=True, missing=None, data_key="dicountCodeId"
    )
    valid_from = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validFrom"
    )
    valid_until = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validUntil"
    )
    validity_check_time = marshmallow.fields.DateTime(
        allow_none=True, missing=None, data_key="validityCheckTime"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DiscountCodeNonApplicableError(**data)


class DuplicateAttributeValueErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicateAttributeValueError`."""

    attribute = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicateAttributeValueError(**data)


class DuplicateAttributeValuesErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicateAttributeValuesError`."""

    attributes = helpers.LazyNestedField(
        nested="commercetools._schemas._product.AttributeSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicateAttributeValuesError(**data)


class DuplicateFieldErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicateFieldError`."""

    field = marshmallow.fields.String(allow_none=True, missing=None)
    duplicate_value = marshmallow.fields.Raw(
        allow_none=True, missing=None, data_key="duplicateValue"
    )
    conflicting_resource = helpers.Discriminator(
        discriminator_field=("typeId", "type_id"),
        discriminator_schemas={
            "cart-discount": "commercetools._schemas._cart_discount.CartDiscountReferenceSchema",
            "cart": "commercetools._schemas._cart.CartReferenceSchema",
            "category": "commercetools._schemas._category.CategoryReferenceSchema",
            "channel": "commercetools._schemas._channel.ChannelReferenceSchema",
            "key-value-document": "commercetools._schemas._custom_object.CustomObjectReferenceSchema",
            "customer-group": "commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
            "customer": "commercetools._schemas._customer.CustomerReferenceSchema",
            "discount-code": "commercetools._schemas._discount_code.DiscountCodeReferenceSchema",
            "inventory-entry": "commercetools._schemas._inventory.InventoryEntryReferenceSchema",
            "order-edit": "commercetools._schemas._order_edit.OrderEditReferenceSchema",
            "order": "commercetools._schemas._order.OrderReferenceSchema",
            "payment": "commercetools._schemas._payment.PaymentReferenceSchema",
            "product-discount": "commercetools._schemas._product_discount.ProductDiscountReferenceSchema",
            "product-type": "commercetools._schemas._product_type.ProductTypeReferenceSchema",
            "product": "commercetools._schemas._product.ProductReferenceSchema",
            "review": "commercetools._schemas._review.ReviewReferenceSchema",
            "shipping-method": "commercetools._schemas._shipping_method.ShippingMethodReferenceSchema",
            "shopping-list": "commercetools._schemas._shopping_list.ShoppingListReferenceSchema",
            "state": "commercetools._schemas._state.StateReferenceSchema",
            "store": "commercetools._schemas._store.StoreReferenceSchema",
            "tax-category": "commercetools._schemas._tax_category.TaxCategoryReferenceSchema",
            "type": "commercetools._schemas._type.TypeReferenceSchema",
            "zone": "commercetools._schemas._zone.ZoneReferenceSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="conflictingResource",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicateFieldError(**data)


class DuplicateFieldWithConflictingResourceErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicateFieldWithConflictingResourceError`."""

    field = marshmallow.fields.String(allow_none=True)
    duplicate_value = marshmallow.fields.Raw(allow_none=True, data_key="duplicateValue")
    conflicting_resource = helpers.Discriminator(
        discriminator_field=("typeId", "type_id"),
        discriminator_schemas={
            "cart-discount": "commercetools._schemas._cart_discount.CartDiscountReferenceSchema",
            "cart": "commercetools._schemas._cart.CartReferenceSchema",
            "category": "commercetools._schemas._category.CategoryReferenceSchema",
            "channel": "commercetools._schemas._channel.ChannelReferenceSchema",
            "key-value-document": "commercetools._schemas._custom_object.CustomObjectReferenceSchema",
            "customer-group": "commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
            "customer": "commercetools._schemas._customer.CustomerReferenceSchema",
            "discount-code": "commercetools._schemas._discount_code.DiscountCodeReferenceSchema",
            "inventory-entry": "commercetools._schemas._inventory.InventoryEntryReferenceSchema",
            "order-edit": "commercetools._schemas._order_edit.OrderEditReferenceSchema",
            "order": "commercetools._schemas._order.OrderReferenceSchema",
            "payment": "commercetools._schemas._payment.PaymentReferenceSchema",
            "product-discount": "commercetools._schemas._product_discount.ProductDiscountReferenceSchema",
            "product-type": "commercetools._schemas._product_type.ProductTypeReferenceSchema",
            "product": "commercetools._schemas._product.ProductReferenceSchema",
            "review": "commercetools._schemas._review.ReviewReferenceSchema",
            "shipping-method": "commercetools._schemas._shipping_method.ShippingMethodReferenceSchema",
            "shopping-list": "commercetools._schemas._shopping_list.ShoppingListReferenceSchema",
            "state": "commercetools._schemas._state.StateReferenceSchema",
            "store": "commercetools._schemas._store.StoreReferenceSchema",
            "tax-category": "commercetools._schemas._tax_category.TaxCategoryReferenceSchema",
            "type": "commercetools._schemas._type.TypeReferenceSchema",
            "zone": "commercetools._schemas._zone.ZoneReferenceSchema",
        },
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="conflictingResource",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicateFieldWithConflictingResourceError(**data)


class DuplicatePriceScopeErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicatePriceScopeError`."""

    conflicting_prices = helpers.LazyNestedField(
        nested="commercetools._schemas._common.PriceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        many=True,
        data_key="conflictingPrices",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicatePriceScopeError(**data)


class DuplicateVariantValuesErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.DuplicateVariantValuesError`."""

    variant_values = helpers.LazyNestedField(
        nested="commercetools._schemas._error.VariantValuesSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="variantValues",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.DuplicateVariantValuesError(**data)


class EnumValueIsUsedErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.EnumValueIsUsedError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.EnumValueIsUsedError(**data)


class ExtensionBadResponseErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ExtensionBadResponseError`."""

    localized_message = LocalizedStringField(
        allow_none=True, missing=None, data_key="localizedMessage"
    )
    extension_extra_info = marshmallow.fields.Dict(
        allow_none=True, missing=None, data_key="extensionExtraInfo"
    )
    error_by_extension = helpers.LazyNestedField(
        nested="commercetools._schemas._error.ErrorByExtensionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="errorByExtension",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ExtensionBadResponseError(**data)


class ExtensionNoResponseErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ExtensionNoResponseError`."""

    localized_message = LocalizedStringField(
        allow_none=True, missing=None, data_key="localizedMessage"
    )
    extension_extra_info = marshmallow.fields.Dict(
        allow_none=True, missing=None, data_key="extensionExtraInfo"
    )
    error_by_extension = helpers.LazyNestedField(
        nested="commercetools._schemas._error.ErrorByExtensionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="errorByExtension",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ExtensionNoResponseError(**data)


class ExtensionUpdateActionsFailedErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ExtensionUpdateActionsFailedError`."""

    localized_message = LocalizedStringField(
        allow_none=True, missing=None, data_key="localizedMessage"
    )
    extension_extra_info = marshmallow.fields.Dict(
        allow_none=True, missing=None, data_key="extensionExtraInfo"
    )
    error_by_extension = helpers.LazyNestedField(
        nested="commercetools._schemas._error.ErrorByExtensionSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        data_key="errorByExtension",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ExtensionUpdateActionsFailedError(**data)


class InsufficientScopeErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InsufficientScopeError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InsufficientScopeError(**data)


class InvalidCredentialsErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidCredentialsError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidCredentialsError(**data)


class InvalidCurrentPasswordErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidCurrentPasswordError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidCurrentPasswordError(**data)


class InvalidFieldErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidFieldError`."""

    field = marshmallow.fields.String(allow_none=True)
    invalid_value = marshmallow.fields.Raw(allow_none=True, data_key="invalidValue")
    allowed_values = marshmallow.fields.List(
        marshmallow.fields.Raw(allow_none=True),
        allow_none=True,
        missing=None,
        data_key="allowedValues",
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidFieldError(**data)


class InvalidInputErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidInputError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidInputError(**data)


class InvalidItemShippingDetailsErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidItemShippingDetailsError`."""

    subject = marshmallow.fields.String(allow_none=True)
    item_id = marshmallow.fields.String(allow_none=True, data_key="itemId")

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidItemShippingDetailsError(**data)


class InvalidJsonInputErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidJsonInputError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidJsonInputError(**data)


class InvalidOperationErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidOperationError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidOperationError(**data)


class InvalidSubjectErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidSubjectError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidSubjectError(**data)


class InvalidTokenErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.InvalidTokenError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.InvalidTokenError(**data)


class LanguageUsedInStoresErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.LanguageUsedInStoresError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.LanguageUsedInStoresError(**data)


class MatchingPriceNotFoundErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.MatchingPriceNotFoundError`."""

    product_id = marshmallow.fields.String(allow_none=True, data_key="productId")
    variant_id = marshmallow.fields.Integer(allow_none=True, data_key="variantId")
    currency = marshmallow.fields.String(allow_none=True, missing=None)
    country = marshmallow.fields.String(allow_none=True, missing=None)
    customer_group = helpers.LazyNestedField(
        nested="commercetools._schemas._customer_group.CustomerGroupReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
        data_key="customerGroup",
    )
    channel = helpers.LazyNestedField(
        nested="commercetools._schemas._channel.ChannelReferenceSchema",
        unknown=marshmallow.EXCLUDE,
        allow_none=True,
        missing=None,
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.MatchingPriceNotFoundError(**data)


class MissingTaxRateForCountryErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.MissingTaxRateForCountryError`."""

    tax_category_id = marshmallow.fields.String(
        allow_none=True, data_key="taxCategoryId"
    )
    country = marshmallow.fields.String(allow_none=True, missing=None)
    state = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.MissingTaxRateForCountryError(**data)


class NoMatchingProductDiscountFoundErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.NoMatchingProductDiscountFoundError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.NoMatchingProductDiscountFoundError(**data)


class OutOfStockErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.OutOfStockError`."""

    line_items = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), data_key="lineItems"
    )
    skus = marshmallow.fields.List(marshmallow.fields.String(allow_none=True))

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.OutOfStockError(**data)


class PriceChangedErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.PriceChangedError`."""

    line_items = marshmallow.fields.List(
        marshmallow.fields.String(allow_none=True), data_key="lineItems"
    )
    shipping = marshmallow.fields.Bool(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.PriceChangedError(**data)


class ReferenceExistsErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ReferenceExistsError`."""

    referenced_by = marshmallow_enum.EnumField(
        types.ReferenceTypeId, by_value=True, missing=None, data_key="referencedBy"
    )

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ReferenceExistsError(**data)


class ReferencedResourceNotFoundErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ReferencedResourceNotFoundError`."""

    type_id = marshmallow_enum.EnumField(
        types.ReferenceTypeId, by_value=True, data_key="typeId"
    )
    id = marshmallow.fields.String(allow_none=True, missing=None)
    key = marshmallow.fields.String(allow_none=True, missing=None)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ReferencedResourceNotFoundError(**data)


class RequiredFieldErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.RequiredFieldError`."""

    field = marshmallow.fields.String(allow_none=True)

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.RequiredFieldError(**data)


class ResourceNotFoundErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ResourceNotFoundError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ResourceNotFoundError(**data)


class ShippingMethodDoesNotMatchCartErrorSchema(ErrorObjectSchema):
    """Marshmallow schema for :class:`commercetools.types.ShippingMethodDoesNotMatchCartError`."""

    class Meta:
        unknown = marshmallow.EXCLUDE

    @marshmallow.post_load
    def post_load(self, data, **kwargs):
        del data["code"]
        return types.ShippingMethodDoesNotMatchCartError(**data)



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._message import (
    MessagePagedQueryResponseSchema,
    MessageSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._message import Message, MessagePagedQueryResponse
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _MessageQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class MessageService(abstract.AbstractService):
    """A message represents a change or an action performed on a resource (like an
    Order or a Product)."""

    def get_by_id(self, id: str, *, expand: OptionalListStr = None) -> Message:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"messages/{id}", params=params, schema_cls=MessageSchema
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> MessagePagedQueryResponse:
        """A message represents a change or an action performed on a resource (like
        an Order or a Product).
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _MessageQuerySchema,
        )
        return self._client._get(
            endpoint="messages",
            params=params,
            schema_cls=MessagePagedQueryResponseSchema,
        )



import pytest

from commercetools.utils import fix_token_url


@pytest.mark.parametrize(
    "token_url,expected_url",
    [
        ("https://auth.sphere.io", "https://auth.sphere.io/oauth/token"),
        ("https://auth.sphere.io/oauth/token", "https://auth.sphere.io/oauth/token"),
        ("https://auth.commercetools.co", "https://auth.commercetools.co/oauth/token"),
        (
            "https://auth.sphere.io?test=123",
            "https://auth.sphere.io/oauth/token?test=123",
        ),
    ],
)
def test_fix_token_url(token_url, expected_url):
    assert fix_token_url(token_url) == expected_url



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._zone import (
    LocationSchema,
    ZoneDraftSchema,
    ZonePagedQueryResponseSchema,
    ZoneSchema,
    ZoneUpdateSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import (
    update_attribute,
    update_attribute_add_item,
    update_attribute_delete_item,
)


class ZonesModel(BaseModel):
    _primary_type_name = "zones"
    _resource_schema = ZoneSchema
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.ZoneDraft, id: typing.Optional[str] = None
    ) -> types.Zone:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.Zone(
            id=str(object_id),
            version=1,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            key=draft.key,
            name=draft.name,
            description=draft.description,
            locations=draft.locations,
        )


class ZonesBackend(ServiceBackend):
    service_path = "zones"
    model_class = ZonesModel
    _schema_draft = ZoneDraftSchema
    _schema_update = ZoneUpdateSchema
    _schema_query_response = ZonePagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
        ]

    _actions = {
        "changeName": update_attribute("name", "name"),
        "setKey": update_attribute("key", "key"),
        "setDescription": update_attribute("description", "description"),
        "addLocation": update_attribute_add_item(
            "locations", "location", LocationSchema
        ),
        "removeLocation": update_attribute_delete_item(
            "locations", "location", LocationSchema
        ),
    }



import pytest

from commercetools import types


def test_state_flow(client, state_draft):
    state = client.states.create(state_draft)
    assert state.id

    new_name = types.LocalizedString({"en": "new_name"})
    new_description = types.LocalizedString({"en": "new_description"})
    update_actions = [
        types.StateSetNameAction(name=new_name),
        types.StateSetDescriptionAction(description=new_description),
        types.StateChangeInitialAction(initial=True),
        types.StateSetRolesAction(
            roles=[types.StateRoleEnum.REVIEW_INCLUDED_IN_STATISTICS]
        ),
    ]
    state = client.states.update_by_id(state.id, state.version, update_actions)
    assert state.name == new_name
    assert state.description == new_description
    assert state.initial is True
    assert len(state.roles) == 1

    state = client.states.update_by_id(
        state.id, state.version, [types.StateSetRolesAction(roles=[])]
    )
    assert len(state.roles) == 0

    deleted_state = client.states.delete_by_id(state.id, state.version)
    assert state.id == deleted_state.id


@pytest.fixture
def state_draft():
    return types.StateDraft(
        key="test state",
        name=types.LocalizedString({"en": "test store"}),
        description=types.LocalizedString({"en": "test store"}),
        initial=False,
        type=types.StateTypeEnum.REVIEW_STATE,
    )



# DO NOT EDIT! This file is automatically generated
import marshmallow
from marshmallow import fields

from commercetools.helpers import OptionalList, RemoveEmptyValuesMixin


class ExpandableSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    expand = OptionalList(fields.String(), required=False)


class QuerySchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    where = OptionalList(fields.String(), required=False)
    predicate_var = fields.Dict()

    @marshmallow.post_dump
    def _predicate_var_post_dump(self, data, **kwrags):
        values = data.pop("predicate_var")
        if not values:
            return data
        for key, val in values.items():
            data[f"var.{key}"] = val
        return data

    @marshmallow.pre_load
    def _predicate_var_post_load(self, data, **kwrags):
        items = {}
        for key in list(data.keys()):
            if key.startswith("var."):
                items[key[4:]] = data[key]
                del data[key]
        data["predicate_var"] = items
        return data


class SortableSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    sort = OptionalList(fields.String(), required=False)


class PagingSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    limit = fields.Int(required=False)
    offset = fields.Int(required=False)
    with_total = fields.Bool(data_key="withTotal", required=False, missing=False)


class VersionedSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    version = fields.Int()


class PriceSelectingSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    price_currency = OptionalList(
        fields.String(), data_key="priceCurrency", required=False
    )
    price_country = OptionalList(
        fields.String(), data_key="priceCountry", required=False
    )
    price_customer_group = OptionalList(
        fields.String(), data_key="priceCustomerGroup", required=False
    )
    price_channel = OptionalList(
        fields.String(), data_key="priceChannel", required=False
    )
    locale_projection = OptionalList(
        fields.String(), data_key="localeProjection", required=False
    )
    store_projection = OptionalList(
        fields.String(), data_key="storeProjection", required=False
    )


class DataErasureSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    data_erasure = fields.Bool(data_key="dataErasure", required=False, missing=False)



# DO NOT EDIT! This file is automatically generated
import typing

from marshmallow import fields

from commercetools._schemas._custom_object import (
    CustomObjectDraftSchema,
    CustomObjectPagedQueryResponseSchema,
    CustomObjectSchema,
)
from commercetools.helpers import OptionalList, RemoveEmptyValuesMixin
from commercetools.types._custom_object import (
    CustomObject,
    CustomObjectDraft,
    CustomObjectPagedQueryResponse,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _CustomObjectQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class _CustomObjectDeleteSchema(
    traits.DataErasureSchema, traits.VersionedSchema, traits.ExpandableSchema
):
    version = OptionalList(fields.String(), required=False)


class CustomObjectService(abstract.AbstractService):
    """Store custom JSON values."""

    def get_by_container_and_key(
        self, container, key, *, expand: OptionalListStr = None
    ) -> CustomObject:
        """Get CustomObject by container and key"""
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"custom-objects/{container}/{key}",
            params=params,
            schema_cls=CustomObjectSchema,
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> CustomObjectPagedQueryResponse:
        """The query endpoint allows to retrieve custom objects in a specific
        container or all custom objects.

        For performance reasons, it is highly advisable to query only for custom
        objects in a container by using the container field in the where
        predicate.   Store custom JSON values.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _CustomObjectQuerySchema,
        )
        return self._client._get(
            endpoint="custom-objects",
            params=params,
            schema_cls=CustomObjectPagedQueryResponseSchema,
        )

    def query_by_container(
        self,
        container: str,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> CustomObjectPagedQueryResponse:
        """Store custom JSON values.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _CustomObjectQuerySchema,
        )
        return self._client._get(
            endpoint=f"custom-objects/{container}",
            params=params,
            schema_cls=CustomObjectPagedQueryResponseSchema,
        )

    def create_or_update(
        self, draft: CustomObjectDraft, *, expand: OptionalListStr = None
    ) -> CustomObject:
        """Creates a new custom object or updates an existing custom object.

        If an object with the given container/key exists, the object will be
        replaced with the new value and the version is incremented. If the
        request contains a version and an object with the given container/key
        exists then the version must match the version of the existing object.
        Concurrent updates for the same custom object still can result in a
        Conflict (409) even if the version is not provided. Fields with null
        values will not be saved.   Store custom JSON values.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="custom-objects",
            params=params,
            data_object=draft,
            request_schema_cls=CustomObjectDraftSchema,
            response_schema_cls=CustomObjectSchema,
        )

    def create(
        self, draft: CustomObjectDraft, *, expand: OptionalListStr = None
    ) -> CustomObject:
        """Creates a new custom object or updates an existing custom object.

        If an object with the given container/key exists, the object will be
        replaced with the new value and the version is incremented. If the
        request contains a version and an object with the given container/key
        exists then the version must match the version of the existing object.
        Concurrent updates for the same custom object still can result in a
        Conflict (409) even if the version is not provided. Fields with null
        values will not be saved.   Store custom JSON values.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="custom-objects",
            params=params,
            data_object=draft,
            request_schema_cls=CustomObjectDraftSchema,
            response_schema_cls=CustomObjectSchema,
        )

    def delete_by_container_and_key(
        self,
        container,
        key,
        *,
        data_erasure: bool = None,
        version: str = None,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> CustomObject:
        """Delete CustomObject by container and key"""
        params = self._serialize_params(
            {"data_erasure": data_erasure, "version": version, "expand": expand},
            _CustomObjectDeleteSchema,
        )
        return self._client._delete(
            endpoint=f"custom-objects/{container}/{key}",
            params=params,
            response_schema_cls=CustomObjectSchema,
            force_delete=force_delete,
        )



import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._channel import (
    ChannelDraftSchema,
    ChannelPagedQueryResponseSchema,
    ChannelSchema,
    ChannelUpdateSchema,
)
from commercetools.testing import utils
from commercetools.testing.abstract import BaseModel, ServiceBackend


class ChannelsModel(BaseModel):
    _resource_schema = ChannelSchema
    _primary_type_name = "channel"
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.ChannelDraft, id: typing.Optional[str] = None
    ) -> types.Channel:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.Channel(
            id=str(object_id),
            version=1,
            name=draft.name,
            description=draft.description,
            roles=draft.roles,
            key=draft.key,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            custom=utils.create_from_draft(draft.custom),
        )


class ChannelsBackend(ServiceBackend):
    service_path = "channels"
    model_class = ChannelsModel
    _schema_draft = ChannelDraftSchema
    _schema_update = ChannelUpdateSchema
    _schema_query_response = ChannelPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]



import copy
import importlib
import json
import typing
import uuid

from marshmallow import Schema
from requests_mock import create_response
from requests_mock.request import _RequestObjectProxy

from commercetools import types
from commercetools._schemas._product import ProductSchema
from commercetools.constants import HEADER_CORRELATION_ID
from commercetools.types import CartSetCustomFieldAction, OrderSetCustomFieldAction
from commercetools.types._abstract import _BaseType


class InternalUpdateError(ValueError):
    pass


def parse_request_params(schema: Schema, request: _RequestObjectProxy) -> dict:
    params = flatten_multivaluedict(request.qs)
    obj = schema().load(params)
    return obj


def flatten_multivaluedict(data: dict) -> dict:
    result = {}
    for key, value in data.items():
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        result[key] = value
    return result


def get_product_variants(product_catalog_data: dict) -> typing.List[dict]:
    variants = []
    if product_catalog_data.get("masterVariant"):
        variants.append(product_catalog_data["masterVariant"])
    if product_catalog_data.get("variants"):
        variants += product_catalog_data["variants"]
    return variants


def create_from_draft(draft):
    """Utility method to create normal objects out of draft objects.

    This is used for non-resource objects. For the resources themselves (which contain)
    an id, we have the specific implementations of the `BaseModel`.
    """
    if not draft:
        return None

    if isinstance(draft, types.CustomFieldsDraft):
        return types.CustomFields(
            type=types.TypeReference(id=draft.type.id), fields=draft.fields
        )
    if isinstance(draft, types.PriceTierDraft):
        return types.PriceTier(
            minimum_quantity=draft.minimum_quantity,
            value=_money_to_typed(
                types.Money(
                    cent_amount=draft.value.cent_amount,
                    currency_code=draft.value.currency_code,
                )
            ),
        )

    raise ValueError(f"Unsupported type {draft.__class__}")


def custom_fields_from_draft(
    storage, draft: types.CustomFieldsDraft
) -> types.CustomFields:
    return types.CustomFields(
        type=types.TypeReference(
            id=storage.get_by_resource_identifier(draft.type)["id"]
        ),
        fields=draft.fields,
    )


def _money_to_typed(
    money: typing.Optional[types.Money],
) -> typing.Optional[types.TypedMoney]:
    if money is not None:
        return types.TypedMoney(
            cent_amount=money.cent_amount,
            currency_code=money.currency_code,
            type=types.MoneyType.CENT_PRECISION,
            fraction_digits=2,
        )
    return None


def update_attribute(dst: str, src: str):
    def updater(self, obj, action):
        value = getattr(action, src)

        if isinstance(value, _BaseType):
            # FIXME: this stupidly assumes the types and schemas module names
            # are in sync. which might be the case since we auto-generate both,
            # but still...
            pkg_name = value.__class__.__module__.replace(".types.", "._schemas.")
            module = importlib.import_module(pkg_name)
            schema = getattr(module, value.__class__.__name__ + "Schema", None)
            if schema:
                value = schema().dump(value)
        if obj[dst] != value:
            new = copy.deepcopy(obj)
            new[dst] = value
            return new
        return obj

    return updater


def update_enum_attribute(dst: str, src: str):
    def updater(self, obj, action):
        value = getattr(action, src).value
        if obj[dst] != value:
            new = copy.deepcopy(obj)
            new[dst] = value
            return new
        return obj

    return updater


def update_attribute_add_item(
    dst: str,
    src: str,
    schema: Schema,
    id_generator: typing.Optional[typing.Callable] = None,
):
    def updater(self, obj, action):
        value = getattr(action, src)
        if id_generator:
            value.id = id_generator()
        value = schema().dump(value)
        if value not in obj[dst]:
            new = copy.deepcopy(obj)
            new[dst].append(value)
            return new
        else:
            raise InternalUpdateError(
                "Duplicate value %r exists for field %s on %r"
                % (json.dumps(value), dst, obj["id"])
            )
        return obj

    return updater


def update_attribute_delete_item(dst: str, src: str, schema: Schema):
    def updater(self, obj, action):
        value = getattr(action, src)
        value = schema().dump(value)
        if value in obj[dst]:
            new = copy.deepcopy(obj)
            new[dst].remove(value)
            return new
        else:
            raise InternalUpdateError("No item found with id %r" % (obj["id"]))

    return updater


def update_attribute_delete_item_by_id(dst: str, src: str):
    def updater(self, obj, action):
        value = getattr(action, src)

        new = copy.deepcopy(obj)
        for i, item in enumerate(obj[dst]):
            if item["id"] == value:
                del new[dst][i]
                return new

        raise InternalUpdateError("No item found with id %r" % value)

    return updater


def set_custom_field():
    """Set custom field. Note it always sets the type now, instead of type checking the custom field type!"""

    def updater(
        self,
        obj,
        action: typing.Union[OrderSetCustomFieldAction, CartSetCustomFieldAction],
    ):
        name = action.name
        value = action.value

        # real API always increments version, so always apply new value.
        new = copy.deepcopy(obj)
        if not new["custom"]:
            new["custom"] = {"fields": {}}
        new["custom"]["fields"][name] = value
        return new

    return updater


def get_product_from_storage(
    storage: "commercetools.testing.Storage",
    product_id: uuid.UUID = None,
    sku: str = None,
) -> typing.Optional[types.Product]:
    product = None
    product_store = storage._stores["product"]

    if product_id:
        product = ProductSchema().load(product_store[product_id])
    elif sku:
        for product_data in product_store.values():
            try:
                master_data = product_data["masterData"]["current"]
            except KeyError:
                continue

            if master_data and master_data["masterVariant"]["sku"] == sku:
                product = ProductSchema().load(product_data)
    else:
        raise ValueError("SKU or productId is required")

    return product


def create_commercetools_response(request, **kwargs):
    correlation_id = request.headers.get(
        HEADER_CORRELATION_ID, f"projects-{str(uuid.uuid4())}"
    )
    headers = kwargs.pop("headers", {})
    headers.update({HEADER_CORRELATION_ID: correlation_id})
    return create_response(request, headers=headers, **kwargs)



import threading
import urllib.parse

tls = threading.local()


class BaseTokenSaver:
    def get_token(self, client_id, scopes):
        raise NotImplementedError()

    def add_token(self, client_id, scopes, token):
        raise NotImplementedError()

    def _create_token_hash(self, client_id, scopes):
        assert scopes is not None
        return "%s:%s" % (client_id, ";".join(scopes))


class DefaultTokenSaver(BaseTokenSaver):
    @property
    def storage(self):
        items = getattr(tls, "tokens", None)
        if items is None:
            items = {}
            setattr(tls, "tokens", items)
        return items

    def add_token(self, client_id, scopes, token):
        name = self._create_token_hash(client_id, scopes)
        self.storage[name] = token

    def get_token(self, client_id, scopes):
        name = self._create_token_hash(client_id, scopes)
        return self.storage.get(name)

    @classmethod
    def clear_cache(cls):
        items = getattr(tls, "tokens", {})
        items.clear()


def fix_token_url(token_url: str) -> str:
    """
    Ensure the token url has the right format.

    Often clients only pass the base url instead of the complete
    token url, which gets confusing for users.
    """
    parts = urllib.parse.urlparse(token_url)
    if parts.path == "":
        token_url = urllib.parse.urlunparse((*parts[:2], "/oauth/token", *parts[3:]))
    return token_url



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._state import (
    StateDraftSchema,
    StatePagedQueryResponseSchema,
    StateSchema,
    StateUpdateSchema,
)
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._state import (
    State,
    StateDraft,
    StatePagedQueryResponse,
    StateUpdate,
    StateUpdateAction,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _StateQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
):
    pass


class _StateUpdateSchema(traits.ExpandableSchema, traits.VersionedSchema):
    pass


class _StateDeleteSchema(traits.VersionedSchema, traits.ExpandableSchema):
    pass


class StateService(abstract.AbstractService):
    """The commercetools platform allows you to model states of certain objects,
    such as orders, line items, products,

    reviews, and payments in order to define finite state machines reflecting the
    business logic you'd like to implement.
    """

    def get_by_id(self, id: str, *, expand: OptionalListStr = None) -> State:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"states/{id}", params=params, schema_cls=StateSchema
        )

    def get_by_key(self, key: str, *, expand: OptionalListStr = None) -> State:
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._get(
            endpoint=f"states/key={key}", params=params, schema_cls=StateSchema
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
    ) -> StatePagedQueryResponse:
        """The commercetools platform allows you to model states of certain objects,
        such as orders, line items, products, reviews, and payments in order to
        define finite state machines reflecting the business logic you'd like to
        implement.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
            },
            _StateQuerySchema,
        )
        return self._client._get(
            endpoint="states", params=params, schema_cls=StatePagedQueryResponseSchema
        )

    def create(self, draft: StateDraft, *, expand: OptionalListStr = None) -> State:
        """The commercetools platform allows you to model states of certain objects,
        such as orders, line items, products, reviews, and payments in order to
        define finite state machines reflecting the business logic you'd like to
        implement.
        """
        params = self._serialize_params({"expand": expand}, traits.ExpandableSchema)
        return self._client._post(
            endpoint="states",
            params=params,
            data_object=draft,
            request_schema_cls=StateDraftSchema,
            response_schema_cls=StateSchema,
        )

    def update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[StateUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> State:
        params = self._serialize_params({"expand": expand}, _StateUpdateSchema)
        update_action = StateUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"states/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=StateUpdateSchema,
            response_schema_cls=StateSchema,
            force_update=force_update,
        )

    def update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[StateUpdateAction],
        *,
        expand: OptionalListStr = None,
        force_update: bool = False,
    ) -> State:
        params = self._serialize_params({"expand": expand}, _StateUpdateSchema)
        update_action = StateUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"states/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=StateUpdateSchema,
            response_schema_cls=StateSchema,
            force_update=force_update,
        )

    def delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> State:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _StateDeleteSchema
        )
        return self._client._delete(
            endpoint=f"states/{id}",
            params=params,
            response_schema_cls=StateSchema,
            force_delete=force_delete,
        )

    def delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        force_delete: bool = False,
    ) -> State:
        params = self._serialize_params(
            {"version": version, "expand": expand}, _StateDeleteSchema
        )
        return self._client._delete(
            endpoint=f"states/key={key}",
            params=params,
            response_schema_cls=StateSchema,
            force_delete=force_delete,
        )



import copy
import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._cart import (
    CartDraftSchema,
    CartPagedQueryResponseSchema,
    CartSchema,
    CartUpdateSchema,
)
from commercetools._schemas._order import PaymentInfoSchema
from commercetools._schemas._payment import PaymentResourceIdentifierSchema
from commercetools._schemas._product import ProductSchema
from commercetools.testing import utils
from commercetools.testing.abstract import BaseModel, ServiceBackend


class CartsModel(BaseModel):
    _resource_schema = CartSchema
    _primary_type_name = "cart"

    def _create_line_item_from_draft(
        self, draft: types.CartDraft, line_item_draft: types.LineItemDraft
    ) -> types.LineItem:
        line_id = str(uuid.uuid4())
        price = 1000

        product_data = self._storage.get_by_resource_identifier(
            types.ProductResourceIdentifier(id=line_item_draft.product_id)
        )
        product: types.Product = ProductSchema().load(product_data)

        variant = None
        for v in product.master_data.current.variants:
            if v.id == line_item_draft.variant_id:
                variant = v

        return types.LineItem(
            id=line_id,
            name=types.LocalizedString({"en": line_id}),
            product_id=line_item_draft.product_id,
            product_type=product.product_type,
            variant=variant,
            price=types.Price(
                id=str(uuid.uuid4()),
                value=types.CentPrecisionMoney(
                    currency_code=draft.currency, cent_amount=price, fraction_digits=2
                ),
            ),
            taxed_price=types.TaxedItemPrice(
                total_net=types.CentPrecisionMoney(
                    currency_code=draft.currency,
                    cent_amount=price * (line_item_draft.quantity or 0),
                    fraction_digits=2,
                ),
                total_gross=types.CentPrecisionMoney(
                    currency_code=draft.currency,
                    cent_amount=price * (line_item_draft.quantity or 0),
                    fraction_digits=2,
                ),
            ),
            total_price=types.CentPrecisionMoney(
                currency_code=draft.currency,
                cent_amount=price * (line_item_draft.quantity or 0),
                fraction_digits=2,
            ),
            quantity=line_item_draft.quantity,
            discounted_price_per_quantity=[],
            state=[],
            price_mode=types.LineItemPriceMode.PLATFORM,
            line_item_mode=types.LineItemMode.STANDARD,
            custom=utils.create_from_draft(line_item_draft.custom),
        )

    def _create_from_draft(
        self, draft: types.CartDraft, id: typing.Optional[str] = None
    ) -> types.Cart:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        if draft.line_items:
            line_items = [
                self._create_line_item_from_draft(draft, line_item)
                for line_item in draft.line_items
            ]
        else:
            line_items = []

        total_price = None
        taxed_price = None

        if line_items:
            total_price = types.CentPrecisionMoney(
                currency_code=draft.currency,
                cent_amount=sum(
                    line_item.taxed_price.total_gross.cent_amount
                    for line_item in line_items
                    if line_item.taxed_price and line_item.taxed_price.total_gross
                ),
                fraction_digits=2,
            )
            taxed_price = types.TaxedPrice(
                total_net=types.CentPrecisionMoney(
                    currency_code=draft.currency,
                    cent_amount=sum(
                        line_item.taxed_price.total_net.cent_amount
                        for line_item in line_items
                        if line_item.taxed_price and line_item.taxed_price.total_net
                    ),
                    fraction_digits=2,
                ),
                total_gross=types.CentPrecisionMoney(
                    currency_code=draft.currency,
                    cent_amount=sum(
                        line_item.taxed_price.total_gross.cent_amount
                        for line_item in line_items
                        if line_item.taxed_price and line_item.taxed_price.total_gross
                    ),
                    fraction_digits=2,
                ),
                tax_portions=[
                    types.TaxPortion(
                        name="0% VAT",
                        rate=0,
                        amount=types.CentPrecisionMoney(
                            currency_code=draft.currency,
                            cent_amount=0,
                            fraction_digits=2,
                        ),
                    )
                ],
            )

        # Some fields such as itemShippingAddresses are currently missing. See
        # https://docs.commercetools.com/http-api-projects-carts for a complete overview
        return types.Cart(
            id=str(object_id),
            version=1,
            cart_state=types.CartState.ACTIVE,
            customer_id=draft.customer_id,
            customer_email=draft.customer_email,
            customer_group=draft.customer_group,
            anonymous_id=draft.anonymous_id,
            country=draft.country,
            inventory_mode=draft.inventory_mode,
            tax_mode=draft.tax_mode or types.TaxMode.PLATFORM,
            tax_rounding_mode=draft.tax_rounding_mode or types.RoundingMode.HALF_EVEN,
            tax_calculation_mode=draft.tax_calculation_mode
            or types.TaxCalculationMode.LINE_ITEM_LEVEL,
            line_items=line_items,
            custom_line_items=[],
            refused_gifts=[],
            shipping_address=draft.shipping_address,
            billing_address=draft.billing_address,
            locale=draft.locale,
            origin=draft.origin or types.CartOrigin.CUSTOMER,
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
            custom=utils.create_from_draft(draft.custom),
            total_price=total_price,
            taxed_price=taxed_price,
        )


def add_payment():
    def updater(self, obj, action):
        value = getattr(action, "payment")
        value = PaymentResourceIdentifierSchema().dump(value)
        if not obj["paymentInfo"]:
            obj["paymentInfo"] = PaymentInfoSchema().dump(
                types.PaymentInfo(payments=[])
            )
        if value not in obj["paymentInfo"]["payments"]:
            new = copy.deepcopy(obj)
            new["paymentInfo"]["payments"].append(value)
            return new
        return obj

    return updater


class CartsBackend(ServiceBackend):
    service_path = "carts"
    model_class = CartsModel
    _schema_draft = CartDraftSchema
    _schema_update = CartUpdateSchema
    _schema_query_response = CartPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]

    def set_custom_field(self, obj, action: types.CartSetCustomFieldAction):
        if not obj["custom"]:
            raise ValueError(
                "This resource has no custom type set - please use "
                "setCustomType first to set the type of the custom fields"
            )

        name = action.name
        value = action.value

        # real API always increments version, so always apply new value.
        new = copy.deepcopy(obj)
        new["custom"]["fields"][name] = value
        return new

    def set_custom_type(self, obj, action: types.CartSetCustomTypeAction):
        custom_type = self.model._storage.get_by_resource_identifier(action.type)

        # real API always increments version, so always apply new value.
        new = copy.deepcopy(obj)
        new["custom"] = {
            "type": {"id": custom_type["id"], "typeId": "type"},
            "fields": {},
        }
        return new

    _actions = {
        "addPayment": add_payment(),
        "setCustomField": set_custom_field,
        "setCustomType": set_custom_type,
    }



# DO NOT EDIT! This file is automatically generated
import typing

from commercetools._schemas._graph_ql import GraphQLResponseSchema
from commercetools.helpers import RemoveEmptyValuesMixin
from commercetools.types._graph_ql import GraphQLResponse
from commercetools.typing import OptionalListStr

from . import abstract, traits


class GraphqlService(abstract.AbstractService):
    """The commercetools™ platform provides a GraphQL API"""

    def create(self) -> GraphQLResponse:
        """Execute a GraphQL query

        The commercetools™ platform provides a GraphQL API
        """
        params: typing.Dict[str, str] = {}
        return self._client._post(
            endpoint="graphql",
            params=params,
            data_object=None,
            request_schema_cls=None,
            response_schema_cls=GraphQLResponseSchema,
        )



import pytest

from commercetools import types


def test_api_client_flow(client, api_client_draft):
    api_client = client.api_clients.create(api_client_draft)

    assert api_client.id

    deleted_api_client = client.api_clients.delete_by_id(api_client.id)

    assert api_client.id == deleted_api_client.id


@pytest.fixture
def api_client_draft():
    return types.ApiClientDraft(
        name="Test API Client", scope="manage_project:some_project_key"
    )



# DO NOT EDIT! This file is automatically generated
import typing

import marshmallow
from marshmallow import fields

from commercetools._schemas._product import (
    ProductDraftSchema,
    ProductPagedQueryResponseSchema,
    ProductSchema,
    ProductUpdateSchema,
)
from commercetools.helpers import OptionalList, RemoveEmptyValuesMixin
from commercetools.types._product import (
    Product,
    ProductDraft,
    ProductPagedQueryResponse,
    ProductUpdate,
    ProductUpdateAction,
)
from commercetools.typing import OptionalListStr

from . import abstract, traits


class _ProductGetSchema(traits.ExpandableSchema, traits.PriceSelectingSchema):
    pass


class _ProductQuerySchema(
    traits.ExpandableSchema,
    traits.SortableSchema,
    traits.PagingSchema,
    traits.QuerySchema,
    traits.PriceSelectingSchema,
):
    pass


class _ProductCreateSchema(traits.ExpandableSchema, traits.PriceSelectingSchema):
    pass


class _ProductUpdateSchema(
    traits.ExpandableSchema, traits.VersionedSchema, traits.PriceSelectingSchema
):
    pass


class _ProductDeleteSchema(
    traits.VersionedSchema, traits.ExpandableSchema, traits.PriceSelectingSchema
):
    pass


class _ProductImagesSchema(marshmallow.Schema, RemoveEmptyValuesMixin):
    filename = OptionalList(fields.String(), required=False)
    variant = fields.Int(required=False)
    sku = OptionalList(fields.String(), required=False)
    staged = fields.Bool(required=False, missing=False)


class ProductService(abstract.AbstractService):
    """Products are the sellable goods in an e-commerce project on CTP.

    This document explains some design concepts of products on CTP and describes the
    available HTTP APIs for working with them.
    """

    def get_by_id(
        self,
        id: str,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
    ) -> Product:
        """Gets the full representation of a product by ID."""
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductGetSchema,
        )
        return self._client._get(
            endpoint=f"products/{id}", params=params, schema_cls=ProductSchema
        )

    def get_by_key(
        self,
        key: str,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
    ) -> Product:
        """Gets the full representation of a product by Key."""
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductGetSchema,
        )
        return self._client._get(
            endpoint=f"products/key={key}", params=params, schema_cls=ProductSchema
        )

    def query(
        self,
        *,
        expand: OptionalListStr = None,
        sort: OptionalListStr = None,
        limit: int = None,
        offset: int = None,
        with_total: bool = None,
        where: OptionalListStr = None,
        predicate_var: typing.Dict[str, str] = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
    ) -> ProductPagedQueryResponse:
        """You can use the query endpoint to get the full representations of
        products.

        REMARK: We suggest to use the performance optimized search endpoint which
        has a bunch functionalities, the query API lacks like sorting on custom
        attributes, etc.   Products are the sellable goods in an e-commerce
        project on CTP. This document explains some design concepts of products
        on CTP and describes the available HTTP APIs for working with them.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "sort": sort,
                "limit": limit,
                "offset": offset,
                "with_total": with_total,
                "where": where,
                "predicate_var": predicate_var,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductQuerySchema,
        )
        return self._client._get(
            endpoint="products",
            params=params,
            schema_cls=ProductPagedQueryResponseSchema,
        )

    def create(
        self,
        draft: ProductDraft,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
    ) -> Product:
        """To create a new product, send a representation that is going to become
        the initial staged representation

        of the new product in the master catalog. If price selection query
        parameters are provided, the selected prices will be added to the
        response.   Products are the sellable goods in an e-commerce project on
        CTP. This document explains some design concepts of products on CTP and
        describes the available HTTP APIs for working with them.
        """
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductCreateSchema,
        )
        return self._client._post(
            endpoint="products",
            params=params,
            data_object=draft,
            request_schema_cls=ProductDraftSchema,
            response_schema_cls=ProductSchema,
        )

    def update_by_id(
        self,
        id: str,
        version: int,
        actions: typing.List[ProductUpdateAction],
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        force_update: bool = False,
    ) -> Product:
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductUpdateSchema,
        )
        update_action = ProductUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"products/{id}",
            params=params,
            data_object=update_action,
            request_schema_cls=ProductUpdateSchema,
            response_schema_cls=ProductSchema,
            force_update=force_update,
        )

    def update_by_key(
        self,
        key: str,
        version: int,
        actions: typing.List[ProductUpdateAction],
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        force_update: bool = False,
    ) -> Product:
        params = self._serialize_params(
            {
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductUpdateSchema,
        )
        update_action = ProductUpdate(version=version, actions=actions)
        return self._client._post(
            endpoint=f"products/key={key}",
            params=params,
            data_object=update_action,
            request_schema_cls=ProductUpdateSchema,
            response_schema_cls=ProductSchema,
            force_update=force_update,
        )

    def delete_by_id(
        self,
        id: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Product:
        params = self._serialize_params(
            {
                "version": version,
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductDeleteSchema,
        )
        return self._client._delete(
            endpoint=f"products/{id}",
            params=params,
            response_schema_cls=ProductSchema,
            force_delete=force_delete,
        )

    def delete_by_key(
        self,
        key: str,
        version: int,
        *,
        expand: OptionalListStr = None,
        price_currency: OptionalListStr = None,
        price_country: OptionalListStr = None,
        price_customer_group: OptionalListStr = None,
        price_channel: OptionalListStr = None,
        locale_projection: OptionalListStr = None,
        store_projection: OptionalListStr = None,
        force_delete: bool = False,
    ) -> Product:
        params = self._serialize_params(
            {
                "version": version,
                "expand": expand,
                "price_currency": price_currency,
                "price_country": price_country,
                "price_customer_group": price_customer_group,
                "price_channel": price_channel,
                "locale_projection": locale_projection,
                "store_projection": store_projection,
            },
            _ProductDeleteSchema,
        )
        return self._client._delete(
            endpoint=f"products/key={key}",
            params=params,
            response_schema_cls=ProductSchema,
            force_delete=force_delete,
        )

    def file_upload(
        self,
        id: str,
        fh: typing.BinaryIO,
        *,
        filename: str = None,
        variant: int = None,
        sku: str = None,
        staged: bool = None,
    ) -> Product:
        """Uploads a binary image file to a given product variant.

        The supported image formats are JPEG, PNG and GIF.
        """
        params = self._serialize_params(
            {"filename": filename, "variant": variant, "sku": sku, "staged": staged},
            _ProductImagesSchema,
        )
        return self._client._upload(
            endpoint=f"products/{id}/images",
            params=params,
            response_schema_cls=ProductSchema,
            file=fh,
        )

    def upload_image(
        self,
        id: str,
        fh: typing.BinaryIO,
        *,
        filename: str = None,
        variant: int = None,
        sku: str = None,
        staged: bool = None,
    ) -> Product:
        """Uploads a binary image file to a given product variant.

        The supported image formats are JPEG, PNG and GIF.
        """
        params = self._serialize_params(
            {"filename": filename, "variant": variant, "sku": sku, "staged": staged},
            _ProductImagesSchema,
        )
        return self._client._upload(
            endpoint=f"products/{id}/images",
            params=params,
            response_schema_cls=ProductSchema,
            file=fh,
        )



import copy
import re
import typing
import uuid

import marshmallow
from marshmallow import fields
from requests_mock.request import _RequestObjectProxy

from commercetools import CommercetoolsError, types
from commercetools._schemas._common import BaseResourceSchema
from commercetools._schemas._error import DuplicateFieldErrorSchema, ErrorResponseSchema
from commercetools.helpers import OptionalList
from commercetools.services import traits
from commercetools.testing import utils
from commercetools.testing.predicates import PredicateFilter
from commercetools.testing.utils import create_commercetools_response


class BaseModel:
    _resource_schema: typing.Type[BaseResourceSchema]
    _unique_values: typing.List[str] = []

    def __init__(self, storage):
        self._storage = storage
        self.objects = storage.init_store(self._primary_type_name)

        assert (
            self._resource_schema
        ), f"{self.__class__.__name__} has no resource schema defined"

    def add_existing(self, obj):
        """Add an 'existing' object to the storage.

        This can be used to set a certain initial state in your test case.
        """
        return self._store_obj(obj)

    def add(self, draft, id=None):
        """Add a new object to the storage.

        This will take a draft and generate an unique ID.
        """
        obj = self._create_from_draft(draft, id)
        return self._store_obj(obj)

    def _store_obj(self, obj):
        key = self._generate_key(obj)
        document = self._resource_schema().dump(obj)

        for field in self._unique_values:
            # Check if a document already exists with a duplicate unique value
            if document[field] in [o[field] for o in self.objects.values()]:
                value = document[field]
                if not value:
                    continue

                msg = f"A duplicate value '{value}' exists for field '{field}'."
                error = types.DuplicateFieldError(
                    message=msg,
                    field=field,
                    duplicate_value=value,
                    conflicting_resource=None,
                )
                serialized_errors = [DuplicateFieldErrorSchema().dump(error)]
                raise CommercetoolsError(
                    msg,
                    errors=serialized_errors,
                    response=types.ErrorResponse(
                        status_code=400,
                        message=msg,
                        error=error.code,
                        errors=serialized_errors,
                    ),
                )

        self.objects[key] = document
        return document

    def _generate_key(self, obj):
        return uuid.UUID(obj.id)

    def _create_from_draft(self, draft, id=None):
        raise NotImplementedError()

    def query(self, where):
        objects = list(self.objects.values())
        if not where:
            return objects
        where_clause = " AND ".join(where)
        predicate_filter = PredicateFilter(where_clause, schema=self._resource_schema)
        return [obj for obj in objects if predicate_filter.match(obj)]

    def get_by_id(self, id):
        try:
            return self.objects.get(uuid.UUID(id))
        except ValueError:
            return None

    def get_by_key(self, key):
        for obj in self.objects.values():
            if obj["key"] == key:
                return obj

    def delete_by_id(self, id):
        obj = self.objects.pop(uuid.UUID(id))
        return obj

    def delete_by_key(self, key):
        for obj_id, obj in self.objects.items():
            if obj["key"] == key:
                return self.objects.pop(obj_id)

    def save(self, obj):
        assert obj["id"]
        obj["version"] += 1
        key = uuid.UUID(obj["id"])
        self.objects[key] = obj


class BaseBackend:
    path = None
    _actions: typing.Dict[str, typing.Callable] = {}
    model_class: typing.Any = None

    def __init__(self, storage=None, model=None):
        if model:
            self.model = model
        elif self.model_class:
            if storage:
                self.model = self.model_class(storage)
            else:
                self.model = self.model_class()

    def register(self, adapter):
        adapter.add_matcher(self._matcher)

    def _matcher(self, request, skip_port_check=False):
        if not skip_port_check and request.port != 443:
            return

        if request.hostname not in self.hostnames:
            return
            if not request.route_kwargs["path"]:
                request.route_kwargs["path"] = "/"

        match = re.match(self.path_prefix, request.path)
        if match:
            request.kwargs = match.groupdict()
            try:
                request_path = match.groupdict()["path"]
            except KeyError:
                request_path = ""

            for path, method, callback in self.urls():
                path_match = re.match(path, request_path)

                # Call the view
                if path_match and request.method == method:
                    response = callback(request, **path_match.groupdict())
                    if response is None:
                        raise NotImplementedError(
                            "No response returned by %r" % callback
                        )
                    return response
            return create_commercetools_response(request, status_code=404)


class GenericSchema(traits.QuerySchema, traits.SortableSchema, traits.PagingSchema):
    where = OptionalList(fields.String())
    sort = OptionalList(fields.String())
    expand = OptionalList(fields.String())
    limit = fields.Int()
    offset = fields.Int()


class ServiceBackend(BaseBackend):
    hostnames = ["api.sphere.io", "localhost"]
    model_class: typing.Any = None
    _schema_draft: typing.Optional[typing.Type[marshmallow.Schema]] = None
    _schema_update: typing.Optional[typing.Type[marshmallow.Schema]] = None
    _schema_query_response: typing.Optional[typing.Type[marshmallow.Schema]] = None
    _schema_query_params: marshmallow.Schema = GenericSchema
    _query_default_limit = 20

    _verify_version: bool = True

    @property
    def path_prefix(self):
        return f"/(?P<project>[^/]+)/{self.service_path}/?(?P<path>.*)?"

    def add_existing(self, obj):
        return self.model.add_existing(obj)

    def create(self, request):
        obj = self._schema_draft().loads(request.body)
        data = self.model.add(obj)
        expanded_data = self._expand(request, data)
        return create_commercetools_response(request, json=expanded_data)

    def get_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        if obj:
            expanded_data = self._expand(request, obj)
            return create_commercetools_response(request, json=expanded_data)
        return create_commercetools_response(request, status_code=404)

    def get_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        if obj:
            expanded_data = self._expand(request, obj)
            return create_commercetools_response(request, json=expanded_data)
        return create_commercetools_response(request, status_code=404)

    def query(self, request):
        params = utils.parse_request_params(self._schema_query_params, request)
        results = self.model.query(params.get("where"))
        total_count = len(results)
        limit = params.get("limit", self._query_default_limit)
        if limit:
            results = results[:limit]

        if params.get("expand"):
            expanded_results = []
            for result in results:
                expanded_results.append(self._expand(request, result))
            results = expanded_results

        data = {
            "count": len(results),
            "total": total_count,
            "offset": 0,
            "limit": limit,
            "results": self.model._resource_schema().load(results, many=True),
        }
        content = self._schema_query_response().dumps(data)
        return create_commercetools_response(request, text=content)

    def update_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        return self._update(request, obj)

    def update_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        return self._update(request, obj)

    def delete_by_id(self, request, id):
        obj = self.model.get_by_id(id)
        if obj:
            if self._verify_version:
                response = self._validate_resource_version(request, obj)
                if response is not None:
                    return response

            obj = self.model.delete_by_id(id)
            return create_commercetools_response(request, json=obj)
        return create_commercetools_response(request, status_code=404)

    def delete_by_key(self, request, key):
        obj = self.model.get_by_key(key)
        if obj:
            if self._verify_version:
                response = self._validate_resource_version(request, obj)
                if response is not None:
                    return response

            obj = self.model.delete_by_key(key)
            return create_commercetools_response(request, json=obj)
        return create_commercetools_response(request, status_code=404)

    def _expand(self, request: _RequestObjectProxy, raw_obj):
        params = utils.parse_request_params(self._schema_query_params, request)
        if "expand" not in params:
            return raw_obj

        expanded_obj = copy.deepcopy(raw_obj)

        for expand_term in params["expand"]:
            self._expand_obj(expanded_obj, expand_term)

        return expanded_obj

    def _expand_obj(self, obj, expand_term):
        resource_ids_to_expand = self._determine_resource_identifiers(
            [obj], expand_term.split(".")
        )
        for resource_identifier in resource_ids_to_expand:
            if "typeId" in resource_identifier and "id" in resource_identifier:
                try:
                    for document in self.model._storage._stores[
                        resource_identifier["typeId"]
                    ].values():
                        if document["id"] == resource_identifier["id"]:
                            resource_identifier["obj"] = copy.deepcopy(document)
                except KeyError:
                    continue

    def _determine_resource_identifiers(self, resource_id_list, terms):
        term = terms[0]
        multiple = False

        if term.endswith("[*]"):
            multiple = True
            term = term[:-3]

        if term.endswith("]"):
            if "[" not in term:
                return []

            index = int(term.split("[")[1][:-1])
            term = term[:-3]

            try:
                resource_identifiers = [
                    resource_id[term][index]
                    for resource_id in resource_id_list
                    if resource_id[term][index]
                ]
            except (KeyError, IndexError):
                return []
        else:
            try:
                resource_identifiers = [
                    resource_id[term]
                    for resource_id in resource_id_list
                    if resource_id[term]
                ]
            except KeyError:
                return []

        if multiple:
            resource_identifiers = [
                item for resource_id in resource_identifiers for item in resource_id
            ]
        if len(terms) == 1:
            return resource_identifiers

        if len(terms) > 1:
            return self._determine_resource_identifiers(resource_identifiers, terms[1:])

    def _update(self, request, obj):
        if not obj:
            return create_commercetools_response(request, status_code=404)

        update = self._schema_update().load(request.json())
        if update.actions:
            obj, err = self._apply_update_actions(obj, update)
            if err:
                return create_commercetools_response(
                    request, json=err, status_code=err["statusCode"]
                )
        expanded_obj = self._expand(request, obj)
        return create_commercetools_response(request, json=expanded_obj)

    def _validate_resource_version(self, request, obj):
        update_version = self._get_version_from_request(request)
        if update_version != obj["version"]:
            data = self._create_version_error_response(obj["version"])
            return create_commercetools_response(request, json=data, status_code=409)

    def _get_version_from_request(self, request):
        version_data = request.qs.get("version")
        if version_data:
            return int(version_data[0])
        return request.json().get("version")

    def _apply_update_actions(self, obj, update):
        original_obj = obj

        for action in update.actions:
            func = self._actions.get(action.action)
            if not func:
                print("Missing action for", action.action)
                continue
            try:
                obj = func(self, obj, action)
            except utils.InternalUpdateError as exc:
                return None, self._create_data_error_response(str(exc), obj)

        # Save the updated object to the model
        if obj != original_obj:
            if self._verify_version and obj["version"] != update.version:
                return None, self._create_version_error_response(obj["version"])
            self.model.save(obj)

        # Temporary
        elif not self._actions:
            self.model.save(obj)

        return obj, None

    def _create_data_error_response(self, message, obj):
        return ErrorResponseSchema().dump(
            types.ErrorResponse(
                status_code=400,
                message=message,
                errors=[
                    types.ConcurrentModificationError(
                        message=message, current_version=obj["version"]
                    )
                ],
            )
        )

    def _create_version_error_response(self, version):
        return ErrorResponseSchema().dump(
            types.ErrorResponse(
                status_code=409,
                message="Version mismatch. Concurrent modification.",
                errors=[
                    types.ConcurrentModificationError(
                        message="Version mismatch. Concurrent modification.",
                        current_version=version,
                    )
                ],
            )
        )



import pytest

from commercetools import CommercetoolsError, types


def test_store_flow(client, store_draft):
    store = client.stores.create(store_draft)

    assert store.id

    deleted_store = client.stores.delete_by_id(store.id, version=store.version)

    assert store.id == deleted_store.id


@pytest.fixture
def store_draft():
    return types.StoreDraft(
        key="test store", name=types.LocalizedString({"en": "test store"})
    )


def test_update_actions(commercetools_api, client, store_draft):
    store = client.stores.create(store_draft)

    assert store.languages is None

    store = client.stores.update_by_id(
        store.id,
        store.version,
        actions=[types.StoreSetLanguagesAction(languages=["en-US"])],
    )

    assert store.languages == ["en-US"]


def test_channels_are_set(commercetools_api, client, store_draft):
    store = client.stores.create(store_draft)

    assert store.distribution_channels == []

    channel = client.channels.create(
        types.ChannelDraft(
            key="FOO", roles=[types.ChannelRoleEnum.PRODUCT_DISTRIBUTION]
        )
    )

    store = client.stores.update_by_id(
        store.id,
        store.version,
        actions=[
            types.StoresSetDistributionChannelsAction(
                distribution_channels=[types.ChannelResourceIdentifier(key="FOO")]
            )
        ],
    )

    assert store.distribution_channels[0].id == channel.id


def test_channel_errors(commercetools_api, client, store_draft):
    store = client.stores.create(store_draft)

    client.channels.create(
        types.ChannelDraft(key="BAR", roles=[types.ChannelRoleEnum.INVENTORY_SUPPLY])
    )

    with pytest.raises(CommercetoolsError):
        client.stores.update_by_id(
            store.id,
            store.version,
            actions=[
                types.StoresSetDistributionChannelsAction(
                    distribution_channels=[types.ChannelResourceIdentifier(key="BAR")]
                )
            ],
        )


def test_store_channel_create(commercetools_api, client, store_draft):
    channel = client.channels.create(
        types.ChannelDraft(
            key="FOO", roles=[types.ChannelRoleEnum.PRODUCT_DISTRIBUTION]
        )
    )
    store_draft.distribution_channels = [types.ChannelResourceIdentifier(key="FOO")]

    store = client.stores.create(store_draft)

    assert store.distribution_channels[0].id == channel.id



import pytest
import requests_mock
from requests.exceptions import HTTPError

from commercetools import types


def test_product_types_get_by_id(client):
    product_type = client.product_types.create(
        types.ProductTypeDraft(
            key="test-product-type", name="test", description="something"
        )
    )

    assert product_type.id
    assert product_type.key == "test-product-type"

    product_type = client.product_types.get_by_id(product_type.id)
    assert product_type.id
    assert product_type.key == "test-product-type"

    with pytest.raises(HTTPError) as e:
        client.product_types.get_by_id("invalid")


def test_product_types_get_by_key(client):
    product_type = client.product_types.create(
        types.ProductTypeDraft(
            key="test-product-type", name="test", description="something"
        )
    )

    assert product_type.id
    assert product_type.key == "test-product-type"

    product_type = client.product_types.get_by_key("test-product-type")
    assert product_type.id
    assert product_type.key == "test-product-type"

    with pytest.raises(HTTPError) as e:
        client.product_types.get_by_key("invalid")


def test_product_type_query(client):
    product_type = client.product_types.create(
        types.ProductTypeDraft(
            key="test-product-type1", name="test-1", description="something"
        )
    )
    product_type = client.product_types.create(
        types.ProductTypeDraft(
            key="test-product-type2", name="test-2", description="something"
        )
    )

    # single sort query
    result = client.product_types.query(sort="id asc")
    assert len(result.results) == 2
    assert result.total == 2

    # multiple sort queries
    result = client.product_types.query(sort=["id asc", "name asc"])
    assert len(result.results) == 2
    assert result.total == 2


def test_product_update(client):
    """Test the return value of the update methods.

    It doesn't test the actual update itself.
    TODO: See if this is worth testing since we're using a mocking backend
    """
    product_type = client.product_types.create(
        types.ProductTypeDraft(
            key="test-product-type", name="test", description="something"
        )
    )

    assert product_type.id
    assert product_type.key == "test-product-type"

    product_type = client.product_types.update_by_id(
        id=product_type.id, version=product_type.version, actions=[]
    )
    assert product_type.key == "test-product-type"

    product_type = client.product_types.update_by_key(
        key="test-product-type", version=product_type.version, actions=[]
    )
    assert product_type.key == "test-product-type"



import copy
import datetime
import typing
import uuid

from commercetools import types
from commercetools._schemas._state import (
    StateDraftSchema,
    StatePagedQueryResponseSchema,
    StateReferenceSchema,
    StateSchema,
    StateSetRolesActionSchema,
    StateUpdateSchema,
)
from commercetools.testing.abstract import BaseModel, ServiceBackend
from commercetools.testing.utils import update_attribute


class StateModel(BaseModel):
    _resource_schema = StateSchema
    _primary_type_name = "state"
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.StateDraft, id: typing.Optional[str] = None
    ) -> types.State:
        object_id = str(uuid.UUID(id) if id is not None else uuid.uuid4())
        return types.State(
            id=str(object_id),
            version=1,
            key=draft.key,
            type=draft.type,
            name=draft.name,
            description=draft.description,
            built_in=False,
            roles=draft.roles,
            initial=draft.initial or False,
            transitions=create_references_from_resources(self, draft.transitions),
            created_at=datetime.datetime.now(datetime.timezone.utc),
            last_modified_at=datetime.datetime.now(datetime.timezone.utc),
        )


def create_references_from_resources(
    model: StateModel,
    state_transitions: typing.Optional[typing.List[types.StateResourceIdentifier]],
) -> typing.List[types.StateReference]:
    if not state_transitions:
        return []
    references = []
    for state_identifier in state_transitions:
        if state_identifier.id:
            references.append(types.StateReference(id=state_identifier.id))
        if state_identifier.key:
            obj = model.get_by_key(state_identifier.key)
            if obj:
                references.append(types.StateReference(id=obj["id"]))

    return references


def set_roles():
    def updater(self, obj: dict, action: types.StateSetRolesAction):
        roles = getattr(action, "roles")

        if roles:
            roles = StateSetRolesActionSchema().dump(action)["roles"]

        if obj["roles"] != roles:
            new = copy.deepcopy(obj)
            new["roles"] = roles
            return new
        return obj

    return updater


def set_transitions():
    def updater(self, obj: dict, action: types.StateSetTransitionsAction):
        transitions = action.transitions
        references = []
        if action.transitions:
            for reference in create_references_from_resources(self.model, transitions):
                references.append(StateReferenceSchema().dump(reference))

        if obj["transitions"] != references:
            new = copy.deepcopy(obj)
            new["transitions"] = references
            return new
        return obj

    return updater


class StatesBackend(ServiceBackend):
    service_path = "states"
    model_class = StateModel
    _schema_draft = StateDraftSchema
    _schema_update = StateUpdateSchema
    _schema_query_response = StatePagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]

    _actions = {
        "setName": update_attribute("name", "name"),
        "setDescription": update_attribute("description", "description"),
        "changeKey": update_attribute("key", "key"),
        "changeType": update_attribute("type", "type"),
        "changeInitial": update_attribute("initial", "initial"),
        "setRoles": set_roles(),
        "setTransitions": set_transitions(),
    }



import datetime
import uuid
from typing import Optional

from commercetools import types
from commercetools._schemas._customer import (
    CustomerDraftSchema,
    CustomerPagedQueryResponseSchema,
    CustomerSchema,
    CustomerUpdateSchema,
)
from commercetools.testing import abstract, utils
from commercetools.testing.utils import create_commercetools_response


class CustomerModel(abstract.BaseModel):
    _resource_schema = CustomerSchema
    _primary_type_name = "customer"
    _unique_values = ["key"]

    def _create_from_draft(
        self, draft: types.CustomerDraft, id: Optional[str] = None
    ) -> types.Customer:
        object_id = uuid.UUID(id) if id is not None else uuid.uuid4()
        now = datetime.datetime.now(datetime.timezone.utc)

        return types.Customer(
            id=str(object_id),
            version=1,
            created_at=now,
            last_modified_at=now,
            customer_number=draft.customer_number,
            email=draft.email,
            password=draft.password,
            first_name=draft.first_name,
            last_name=draft.last_name,
            middle_name=draft.middle_name,
            title=draft.title,
            date_of_birth=draft.date_of_birth,
            company_name=draft.company_name,
            vat_id=draft.vat_id,
            addresses=draft.addresses,
            default_shipping_address_id=(
                str(draft.default_shipping_address)
                if draft.default_shipping_address
                else None
            ),
            shipping_address_ids=(
                [str(address_id) for address_id in draft.shipping_addresses]
                if draft.shipping_addresses
                else None
            ),
            default_billing_address_id=(
                str(draft.default_billing_address)
                if draft.default_billing_address
                else None
            ),
            billing_address_ids=(
                [str(address_id) for address_id in draft.billing_addresses]
                if draft.billing_addresses
                else None
            ),
            is_email_verified=draft.is_email_verified,
            external_id=draft.external_id,
            customer_group=draft.customer_group,
            custom=utils.create_from_draft(draft.custom),
            locale=draft.locale,
            salutation=draft.salutation,
            key=draft.key,
        )


class CustomerBackend(abstract.ServiceBackend):
    service_path = "customers"
    model_class = CustomerModel
    _schema_draft = CustomerDraftSchema
    _schema_update = CustomerUpdateSchema
    _schema_query_response = CustomerPagedQueryResponseSchema

    def urls(self):
        return [
            ("^$", "GET", self.query),
            ("^$", "POST", self.create),
            ("^key=(?P<key>[^/]+)$", "GET", self.get_by_key),
            ("^key=(?P<key>[^/]+)$", "POST", self.update_by_key),
            ("^key=(?P<key>[^/]+)$", "DELETE", self.delete_by_key),
            ("^(?P<id>[^/]+)$", "GET", self.get_by_id),
            ("^(?P<id>[^/]+)$", "POST", self.update_by_id),
            ("^(?P<id>[^/]+)$", "DELETE", self.delete_by_id),
        ]

    def create(self, request):
        obj = self._schema_draft().loads(request.body)
        data = self.model.add(obj)

        # Convert to CustomerSignInResult
        data = {"customer": data, "cart": None}

        expanded_data = self._expand(request, data)
        return create_commercetools_response(request, json=expanded_data)
